@@ -31,6 +31,7 @@ from bup.helpers import (Sha1, add_error, chunkyreader, debug1, debug2,
                          temp_dir,
                          unlink,
                          utc_offset_str)
+from bup.midx import open_midx
 
 
 verbose = 0
@@ -69,9 +70,10 @@ def _git_exo(cmd, **kwargs):
 
 def git_config_get(option, repo_dir=None, opttype=None, cfg_file=None):
     assert not (repo_dir and cfg_file), "repo_dir and cfg_file cannot both be used"
-    cmd = [b'git', b'--git-dir', repo_dir or repo(), b'config', b'--null']
     if cfg_file:
-        cmd.extend([b'--file', cfg_file])
+        cmd = [b'git', b'config', b'--file', cfg_file, b'--null']
+    else:
+        cmd = [b'git', b'--git-dir', repo_dir or repo(), b'config', b'--null']
     if opttype == 'int':
         cmd.extend([b'--int'])
     elif opttype == 'bool':
@@ -448,6 +450,11 @@ class PackIdxV1(PackIdx):
         for ofs in range(start, start + 24 * self.nsha, 24):
             yield self.map[ofs : ofs + 20]
 
+    def oid_offsets_and_idxs(self):
+        end = self.sha_ofs + self.nsha * 24
+        for i, ofs in enumerate(range(self.sha_ofs, end, 24)):
+            yield struct.unpack_from('!I', self.map, offset=ofs)[0], i
+
     def close(self):
         self.closed = True
         if self.map is not None:
@@ -488,16 +495,19 @@ class PackIdxV2(PackIdx):
     def __len__(self):
         return int(self.nsha)  # int() from long for python 2
 
+    def _oid_ofs_from_ofs32_ofs(self, ofs32_ofs):
+        ofs32 = struct.unpack_from('!I', self.map, offset=ofs32_ofs)[0]
+        if ofs32 & 0x80000000:
+            ofs64_i = ofs32 & 0x7fffffff
+            ofs64_ofs = self.ofs64table_ofs + ofs64_i * 8
+            return struct.unpack_from('!Q', self.map, offset=ofs64_ofs)[0]
+        return ofs32
+
     def _ofs_from_idx(self, idx):
         if idx >= self.nsha or idx < 0:
             raise IndexError('invalid pack index index %d' % idx)
-        ofs_ofs = self.ofstable_ofs + idx * 4
-        ofs = struct.unpack_from('!I', self.map, offset=ofs_ofs)[0]
-        if ofs & 0x80000000:
-            idx64 = ofs & 0x7fffffff
-            ofs64_ofs = self.ofs64table_ofs + idx64 * 8
-            ofs = struct.unpack_from('!Q', self.map, offset=ofs64_ofs)[0]
-        return ofs
+        ofs32_ofs = self.ofstable_ofs + idx * 4
+        return self._oid_ofs_from_ofs32_ofs(ofs32_ofs)
 
     def _idx_to_hash(self, idx):
         if idx >= self.nsha or idx < 0:
@@ -517,6 +527,11 @@ class PackIdxV2(PackIdx):
             self.map.close()
             self.map = None
 
+    def oid_offsets_and_idxs(self):
+        end = self.ofstable_ofs + self.nsha * 4
+        for i, ofs32_ofs in enumerate(range(self.ofstable_ofs, end, 4)):
+            yield self._oid_ofs_from_ofs32_ofs(ofs32_ofs), i
+
     def __del__(self):
         assert self.closed
 
@@ -636,20 +651,20 @@ class PackIdxList:
                             d[os.path.join(self.dir, name)] = ix
                 for full in midxes:
                     if not d.get(full):
-                        mx = midx.PackMidx(full)
-                        (mxd, mxf) = os.path.split(mx.name)
-                        broken = False
-                        for n in mx.idxnames:
-                            if not os.path.exists(os.path.join(mxd, n)):
+                        mx, missing = None, None
+                        try:
+                            mx = open_midx(full, ignore_missing=False)
+                        except midx.MissingIdxs as ex:
+                            missing = ex.paths
+                        if not missing:
+                            if mx: midxl.append(mx)
+                        else:
+                            mxd, mxf = os.path.split(full)
+                            for n in missing:
                                 log(('warning: index %s missing\n'
                                      '  used by %s\n')
                                     % (path_msg(n), path_msg(mxf)))
-                                broken = True
-                        if broken:
-                            mx.close()
                             unlink(full)
-                        else:
-                            midxl.append(mx)
                 midxl.sort(key=lambda ix:
                            (-len(ix), -xstat.stat(ix.name).st_mtime))
                 for ix in midxl:
@@ -664,7 +679,7 @@ class PackIdxList:
                         d[ix.name] = ix
                         for name in ix.idxnames:
                             d[os.path.join(self.dir, name)] = ix
-                    elif not ix.force_keep:
+                    else:
                         debug1('midx: removing redundant: %s\n'
                                % path_msg(os.path.basename(ix.name)))
                         ix.close()
@@ -708,25 +723,35 @@ class PackIdxList:
 
 
 def open_idx(filename):
-    if filename.endswith(b'.idx'):
-        f = open(filename, 'rb')
+    if not filename.endswith(b'.idx'): # why is this enforced *here*?
+        raise GitError('pack idx filenames must end with .idx')
+    f = open(filename, 'rb')
+    with ExitStack() as contexts:
+        contexts.enter_context(f)
         header = f.read(8)
         if header[0:4] == b'\377tOc':
             version = struct.unpack('!I', header[4:8])[0]
             if version == 2:
+                contexts.pop_all()
                 return PackIdxV2(filename, f)
             else:
                 raise GitError('%s: expected idx file version 2, got %d'
                                % (path_msg(filename), version))
         elif len(header) == 8 and header[0:4] < b'\377tOc':
+            contexts.pop_all()
             return PackIdxV1(filename, f)
         else:
             raise GitError('%s: unrecognized idx file header'
                            % path_msg(filename))
+
+
+def open_object_idx(filename):
+    if filename.endswith(b'.idx'):
+        return open_idx(filename)
     elif filename.endswith(b'.midx'):
-        return midx.PackMidx(filename)
+        return open_midx(filename)
     else:
-        raise GitError('idx filenames must end with .idx or .midx')
+        raise GitError('pack index filenames must end with .idx or .midx')
 
 
 def idxmerge(idxlist, final_progress=True):
@@ -1285,33 +1310,65 @@ class CatPipe:
     def __init__(self, repo_dir = None):
         require_suitable_git()
         self.repo_dir = repo_dir
-        self.p = self.inprogress = None
+        self.p = self.pcheck = self.inprogress = None
+
+        # probe for cat-file --batch-command
+        tmp = subprocess.Popen([b'git', b'cat-file', b'--batch-command'],
+                               stdin=subprocess.DEVNULL,
+                               stdout=subprocess.DEVNULL,
+                               stderr=subprocess.DEVNULL,
+                               close_fds=True,
+                               env=_gitenv(self.repo_dir))
+        tmp.wait();
+        self.have_batch_command = tmp.returncode == 0
 
     def close(self, wait=False):
         self.p, p = None, self.p
+        self.pcheck, pcheck = None, self.pcheck
         self.inprogress = None
         if p:
             try:
                 p.stdout.close()
             finally:
-                # This will handle pending exceptions correctly once
-                # we drop py2
                 p.stdin.close()
+        if pcheck and pcheck != p:
+            try:
+                pcheck.stdout.close()
+            finally:
+                pcheck.stdin.close()
         if wait:
-            p.wait()
-            return p.returncode
+            if p: p.wait()
+            if pcheck: pcheck.wait()
+            if p and p.returncode:
+                return p.returncode
+            if pcheck and pcheck.returncode:
+                return pcheck.returncode
+            return 0
         return None
 
     def restart(self):
         self.close()
-        self.p = subprocess.Popen([b'git', b'cat-file', b'--batch'],
+        self.p = subprocess.Popen([b'git', b'cat-file',
+                                  b'--batch-command' if self.have_batch_command else b'--batch'],
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   close_fds = True,
                                   bufsize = 4096,
                                   env=_gitenv(self.repo_dir))
 
-    def get(self, ref):
+    def _open_check(self):
+        if self.pcheck is not None: return
+        if self.have_batch_command:
+            self.pcheck = self.p
+            return
+        self.pcheck = subprocess.Popen([b'git', b'cat-file', b'--batch-check'],
+                                       stdin=subprocess.PIPE,
+                                       stdout=subprocess.PIPE,
+                                       close_fds = True,
+                                       bufsize = 4096,
+                                       env=_gitenv(self.repo_dir))
+
+    def get(self, ref, include_data=True):
         """Yield (oidx, type, size), followed by the data referred to by ref.
         If ref does not exist, only yield (None, None, None).
 
@@ -1328,12 +1385,21 @@ class CatPipe:
         assert ref.find(b'\r') < 0
         assert not ref.startswith(b'-')
         self.inprogress = ref
-        self.p.stdin.write(ref + b'\n')
-        self.p.stdin.flush()
-        hdr = self.p.stdout.readline()
+        if include_data:
+            p = self.p
+            if self.have_batch_command:
+                p.stdin.write(b'contents ')
+        else:
+            self._open_check()
+            p = self.pcheck
+            if self.have_batch_command:
+                p.stdin.write(b'info ')
+        p.stdin.write(ref + b'\n')
+        p.stdin.flush()
+        hdr = p.stdout.readline()
         if not hdr:
             raise GitError('unexpected cat-file EOF (last request: %r, exit: %s)'
-                           % (ref, self.p.poll() or 'none'))
+                           % (ref, p.poll() or 'none'))
         if hdr.endswith(b' missing\n'):
             self.inprogress = None
             yield None, None, None
@@ -1342,37 +1408,43 @@ class CatPipe:
         if len(info) != 3 or len(info[0]) != 40:
             raise GitError('expected object (id, type, size), got %r' % info)
         oidx, typ, size = info
+
+        if not include_data:
+            self.inprogress = None
+            yield oidx, typ, size
+            return
+
         size = int(size)
         try:
-            it = chunkyreader(self.p.stdout, size)
+            it = chunkyreader(p.stdout, size)
             yield oidx, typ, size
-            for blob in chunkyreader(self.p.stdout, size):
+            for blob in chunkyreader(p.stdout, size):
                 yield blob
-            readline_result = self.p.stdout.readline()
+            readline_result = p.stdout.readline()
             assert readline_result == b'\n'
             self.inprogress = None
         except Exception as ex:
             with pending_raise(ex):
                 self.close()
 
-    def _join(self, it):
+    def _join(self, oid, path):
+        it = self.get(oid)
         _, typ, _ = next(it)
         if typ == b'blob':
-            for blob in it:
-                yield blob
+            yield from it
         elif typ == b'tree':
             treefile = b''.join(it)
-            for (mode, name, sha) in tree_decode(treefile):
-                for blob in self.join(hexlify(sha)):
-                    yield blob
+            for mode, name, sha in tree_decode(treefile):
+                yield from self._join(hexlify(sha), path + [name])
         elif typ == b'commit':
             treeline = b''.join(it).split(b'\n')[0]
             assert treeline.startswith(b'tree ')
-            for blob in self.join(treeline[5:]):
-                yield blob
+            yield from self._join(treeline[5:], path + [f'commit:{oid!r}'])
+        elif typ is None:
+            path += [repr(oid)]
+            raise GitError(f'missing ref at {path!r}')
         else:
-            raise GitError('invalid object type %r: expected blob/tree/commit'
-                           % typ)
+            raise GitError(f'ref {oid!r} type {typ!r} is not blob/tree/commit')
 
     def join(self, id):
         """Generate a list of the content of all blobs that can be reached
@@ -1380,8 +1452,7 @@ class CatPipe:
         or a commit. The content of all blobs that can be seen from trees or
         commits will be added to the list.
         """
-        for d in self._join(self.get(id)):
-            yield d
+        yield from self._join(id, [])
 
 
 _cp = {}
@@ -1424,52 +1495,73 @@ class MissingObject(KeyError):
         KeyError.__init__(self, 'object %r is missing' % hexlify(oid))
 
 
-WalkItem = namedtuple('WalkItem', ['oid', 'type', 'mode',
-                                   'path', 'chunk_path', 'data'])
-# The path is the mangled path, and if an item represents a fragment
-# of a chunked file, the chunk_path will be the chunked subtree path
-# for the chunk, i.e. ['', '2d3115e', ...].  The top-level path for a
-# chunked file will have a chunk_path of [''].  So some chunk subtree
-# of the file '/foo/bar/baz' might look like this:
-#
-#   item.path = ['foo', 'bar', 'baz.bup']
-#   item.chunk_path = ['', '2d3115e', '016b097']
-#   item.type = 'tree'
-#   ...
+class WalkItem:
+    # The path is the mangled path, and if an item represents a fragment
+    # of a chunked file, the chunk_path will be the chunked subtree path
+    # for the chunk, i.e. ['', '2d3115e', ...].  The top-level path for a
+    # chunked file will have a chunk_path of [''].  So some chunk subtree
+    # of the file '/foo/bar/baz' might look like this:
+    #
+    #   item.path = ['foo', 'bar', 'baz.bup']
+    #   item.chunk_path = ['', '2d3115e', '016b097']
+    #   item.type = 'tree'
+    #   ...
+    __slots__ = 'oid', 'type', 'mode', 'path', 'chunk_path', 'data'
+    def __init__(self, *, oid, type, mode, path, chunk_path, data):
+        self.oid = oid
+        self.type = type
+        self.mode = mode
+        self.path = path
+        self.chunk_path = chunk_path
+        self.data = data
 
-
-def walk_object(get_ref, oidx, stop_at=None, include_data=None):
-    """Yield everything reachable from oidx via get_ref (which must behave
-    like CatPipe get) as a WalkItem, stopping whenever stop_at(oidx)
-    returns true.  Throw MissingObject if a hash encountered is
-    missing from the repository, and don't read or return blob content
-    in the data field unless include_data is set.
+def walk_object(get_ref, oidx, *, stop_at=None, include_data=None,
+                oid_exists=None):
+    """Yield everything reachable from oidx via get_ref (which must
+    behave like CatPipe get) as a WalkItem, stopping whenever
+    stop_at(oidx) returns logically true.  Set the data field to False
+    when the object is missing, or None if the object exists but
+    include_data is logically false.  Missing blobs may not be noticed
+    unless include_data is logically true or oid_exists(oid) is
+    provided.  Yield items depth first, post-order, i.e. parents after
+    children.
 
     """
+
     # Maintain the pending stack on the heap to avoid stack overflow
-    pending = [(oidx, [], [], None)]
+    pending = [(oidx, [], [], None, None)]
     while len(pending):
-        oidx, parent_path, chunk_path, mode = pending.pop()
-        oid = unhexlify(oidx)
+        if isinstance(pending[-1], WalkItem):
+            yield pending.pop()
+            continue
+
+        oidx, parent_path, chunk_path, mode, exp_typ = pending.pop()
         if stop_at and stop_at(oidx):
             continue
 
-        if (not include_data) and mode and stat.S_ISREG(mode):
+        oid = unhexlify(oidx)
+
+        if (not include_data) and mode and exp_typ == b'blob':
             # If the object is a "regular file", then it's a leaf in
             # the graph, so we can skip reading the data if the caller
             # hasn't requested it.
             yield WalkItem(oid=oid, type=b'blob',
                            chunk_path=chunk_path, path=parent_path,
                            mode=mode,
-                           data=None)
+                           data=bool(oid_exists(oid)) if oid_exists else None)
             continue
 
         item_it = get_ref(oidx)
         get_oidx, typ, _ = next(item_it)
         if not get_oidx:
-            raise MissingObject(unhexlify(oidx))
+            yield WalkItem(oid=unhexlify(oidx), type=exp_typ,
+                           chunk_path=chunk_path, path=parent_path,
+                           mode=mode, data=False)
+            continue
         if typ not in (b'blob', b'commit', b'tree'):
             raise Exception('unexpected repository object type %r' % typ)
+        if exp_typ and typ != exp_typ:
+            raise Exception(f'{oidx.decode("ascii")} object type {typ} != {exp_typ}')
 
         # FIXME: set the mode based on the type when the mode is None
         if typ == b'blob' and not include_data:
@@ -1480,18 +1572,23 @@ def walk_object(get_ref, oidx, stop_at=None, include_data=None):
         else:
             data = b''.join(item_it)
 
-        yield WalkItem(oid=oid, type=typ,
-                       chunk_path=chunk_path, path=parent_path,
-                       mode=mode,
-                       data=(data if include_data else None))
+        item = WalkItem(oid=oid, type=typ,
+                        chunk_path=chunk_path, path=parent_path,
+                        mode=mode,
+                        data=(data if include_data else None))
 
-        if typ == b'commit':
+        if typ == b'blob':
+            yield item
+        elif typ == b'commit':
+            pending.append(item)
             commit_items = parse_commit(data)
+            # For now, all paths are rooted at the "nearest" commit
             for pid in commit_items.parents:
-                pending.append((pid, parent_path, chunk_path, mode))
-            pending.append((commit_items.tree, parent_path, chunk_path,
-                            hashsplit.GIT_MODE_TREE))
+                pending.append((pid, [], [], mode, b'commit'))
+            pending.append((commit_items.tree, [oidx], [],
+                            hashsplit.GIT_MODE_TREE, b'tree'))
         elif typ == b'tree':
+            pending.append(item)
             for mode, name, ent_id in tree_decode(data):
                 demangled, bup_type = demangle_name(name, mode)
                 if chunk_path:
@@ -1504,4 +1601,5 @@ def walk_object(get_ref, oidx, stop_at=None, include_data=None):
                     else:
                         sub_chunk_path = chunk_path
                 pending.append((hexlify(ent_id), sub_path, sub_chunk_path,
-                                mode))
+                                mode,
+                                b'tree' if stat.S_ISDIR(mode) else b'blob'))
